# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class HalfFloatArrayTest < Test::Unit::TestCase
  sub_test_case(".new") do
    test("Float") do
      array = Arrow::HalfFloatArray.new([1.5])
      assert_equal([1.5], array.to_a)
    end

    test("Integer") do
      one_half = Arrow::HalfFloat.new(1.5)
      array = Arrow::HalfFloatArray.new([one_half.to_uint16])
      assert_equal([one_half.to_f], array.to_a)
    end

    test("HalfFloat") do
      one_half = Arrow::HalfFloat.new(1.5)
      array = Arrow::HalfFloatArray.new([one_half])
      assert_equal([one_half.to_f], array.to_a)
    end
  end

  test("#[]") do
    one_half = Arrow::HalfFloat.new(1.5)
    array = Arrow::HalfFloatArray.new([one_half.to_uint16])
    assert_equal(one_half.to_f, array[0])
  end
end
