# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestElementWiseAggregateOptions < Test::Unit::TestCase
  include Helper::Buildable

  def setup
    @options = Arrow::ElementWiseAggregateOptions.new
  end

  def test_skip_nulls_property
    assert do
      @options.skip_nulls?
    end
    @options.skip_nulls = false
    assert do
      !@options.skip_nulls?
    end
  end

  def test_min_element_wise_function_with_skip_nulls_false
    args = [
      Arrow::ArrayDatum.new(build_int64_array([1, nil, 3])),
      Arrow::ArrayDatum.new(build_int64_array([4, 1, 5])),
    ]
    @options.skip_nulls = false
    min_element_wise_function = Arrow::Function.find("min_element_wise")
    assert_equal(build_int64_array([1, nil, 3]),
                 min_element_wise_function.execute(args, @options).value)
  end
end
