# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestExpressionBuilderRecord < Test::Unit::TestCase
  def setup
    @table = Arrow::Table.new(field: Arrow::Int32Array.new([1, 13, 3, 17]))
    @schema = @table.schema
  end

  def build
    record = Gandiva::ExpressionBuilder::Record.new(@schema)
    builder = yield(record)
    builder.build
  end

  test("name") do
    node = build do |record|
      record.field
    end
    assert_equal("(int32) field",
                 node.to_s)
  end

  test("#[]") do
    node = build do |record|
      record[:field]
    end
    assert_equal("(int32) field",
                 node.to_s)
  end
end
