{#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#}
*<https://s3.amazonaws.com/arrow-data/index.html|Archery crossbow report> for <{{ report.url(report.job.branch) }}|{{ report.job.branch }}>*
{% if report.tasks_by_state["failure"] %}
:x: *{{ report.tasks_by_state["failure"] | length }} failed jobs*
{% for task_name, task in report.tasks_by_state["failure"] | dictsort -%}
- <{{ report.task_url(task) }}|{{ task_name }}>
{% endfor %}
{%- endif -%}
{% if report.tasks_by_state["error"] %}
:x: *{{ report.tasks_by_state["error"] | length }} errored jobs*
{% for task_name, task in report.tasks_by_state["error"] | dictsort -%}
- <{{ report.task_url(task) }}|{{ task_name }}>
{% endfor %}
{%- endif -%}
{% if report.tasks_by_state["pending"] %}
:warning: *{{ report.tasks_by_state["pending"] | length }} pending jobs*
{%- endif -%}
{% if report.tasks_by_state["success"] %}

:tada: *{{ report.tasks_by_state["success"] | length }} successful jobs*
{%- endif -%}

{% if extra_message_success and not report.contains_failures %}

{{ extra_message_success }}
{%- endif -%}
{% if extra_message_failure and report.contains_failures %}

{{ extra_message_failure }}
{% endif %}