{#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
-#}
Arrow Build Report for Job {{ report.job.branch }}

See https://s3.amazonaws.com/arrow-data/index.html for more information.

All tasks: {{ report.url(report.job.branch) }}
{% if report.tasks_by_state["failure"] %}
Failed Tasks:

{% for task_name, task in report.tasks_by_state["failure"] | dictsort -%}
- {{ task_name }}
  {{ report.task_url(task) }}
{% endfor %}
{% endif %}
{%- if report.tasks_by_state["error"] -%}
Errored Tasks:

{% for task_name, task in report.tasks_by_state["error"] | dictsort -%}
- {{ task_name }}
  {{ report.task_url(task) }}
{% endfor %}
{% endif %}
{%- if report.tasks_by_state["pending"] -%}
Pending Tasks:

{% for task_name, task in report.tasks_by_state["pending"] | dictsort -%}
- {{ task_name }}
  {{ report.task_url(task) }}
{% endfor %}
{% endif %}
{%- if report.tasks_by_state["success"] -%}
Succeeded Tasks:

{% for task_name, task in report.tasks_by_state["success"] | dictsort -%}
- {{ task_name }}
  {{ report.task_url(task) }}
{% endfor %}
{%- endif -%}
