%TNUMERIC Unit tests for arrow.internal.validate.numeric.

% Licensed to the Apache Software Foundation (ASF) under one or more
% contributor license agreements.  See the NOTICE file distributed with
% this work for additional information regarding copyright ownership.
% The ASF licenses this file to you under the Apache License, Version
% 2.0 (the "License"); you may not use this file except in compliance
% with the License.  You may obtain a copy of the License at
%
%   http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
% implied.  See the License for the specific language governing
% permissions and limitations under the License.

classdef tNumeric < matlab.unittest.TestCase
    
    methods(Test)
        
        function ErrorIfWrongType(testCase)
            data = [1 2 2 4];
            fcn = @() arrow.internal.validate.numeric(data, "single");
            errid = "arrow:array:InvalidType";            
            testCase.verifyError(fcn, errid);
        end

        function ErrorIfNonVector(testCase)
            data = [1 2; 2 4];
            fcn = @() arrow.internal.validate.numeric(data, "double");
            errid = "arrow:array:InvalidShape";            
            testCase.verifyError(fcn, errid);
        end

        function ErrorIfNonsparse(testCase)
            data = sparse([1 1 2 2]);
            fcn = @() arrow.internal.validate.numeric(data, "double");
            errid = "arrow:array:Sparse";            
            testCase.verifyError(fcn, errid);
        end

        function ErrorIfComplex(testCase)
            data = [10 + 3i 11 + 4i];
            fcn = @() arrow.internal.validate.numeric(data, "double");
            errid = "arrow:array:ComplexNumeric";            
            testCase.verifyError(fcn, errid);
        end

        function NoErrorIfRealNonsparseNumericVector(testCase)
            data = [10 11 12 13];
            fcn = @() arrow.internal.validate.numeric(data, "double");
            testCase.verifyWarningFree(fcn);
        end
    end
end