%TRECORDBATCH Test class for arrow.tabular.RecordBatch

% Licensed to the Apache Software Foundation (ASF) under one or more
% contributor license agreements.  See the NOTICE file distributed with
% this work for additional information regarding copyright ownership.
% The ASF licenses this file to you under the Apache License, Version
% 2.0 (the "License"); you may not use this file except in compliance
% with the License.  You may obtain a copy of the License at
%
%   http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
% implied.  See the License for the specific language governing
% permissions and limitations under the License.

classdef tRecordBatch < hTabular

    properties
        FromArraysFcn = @arrow.tabular.RecordBatch.fromArrays
        ConstructionFcn = @arrow.recordBatch
        ClassName = "arrow.tabular.RecordBatch"
    end

    methods

        function verifyTabularObject(tc, arrowTabularObj, columnNames, columnTraits, matlabTable)
            tc.verifyInstanceOf(arrowTabularObj, tc.ClassName);
            tc.verifyEqual(arrowTabularObj.NumColumns, int32(width(matlabTable)));
            tc.verifyEqual(arrowTabularObj.ColumnNames, columnNames);
            convertedTable = arrowTabularObj.table();
            tc.verifyEqual(convertedTable, matlabTable);
            for ii = 1:arrowTabularObj.NumColumns
                column = arrowTabularObj.column(ii);
                tc.verifyEqual(column.toMATLAB(), matlabTable{:, ii});
                tc.verifyInstanceOf(column, columnTraits{ii}.ArrayClassName);
             end
        end
        
        function col = makeColumnFromArray(~, array)
            col = array;
        end

    end

end
