# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestIndexOptions < Test::Unit::TestCase
  def setup
    @value = Arrow::Int32Scalar.new(29)
    @options = Arrow::IndexOptions.new
    @options.value = @value
  end

  def test_value
    assert_equal(@value, @options.value)
    new_value = Arrow::UInt8Scalar.new(1)
    @options.value = new_value
    assert_equal(new_value, @options.value)
  end
end
