# shellcheck shell=bash
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# NOTE: this script is not marked executable as it should be source'd
# for `ulimit` to take effect.

set -e

platform=$(uname)

if [ "${platform}" = "Linux" ]; then
  # We need to override `core_pattern` because
  # 1. the original setting may reference apport, which is not available under
  #    most Docker containers;
  # 2. we want to write the core file in a well-known directory.
  sudo sysctl -w kernel.core_pattern="/tmp/core.%e.%p"
fi

ulimit -c unlimited
