% Licensed to the Apache Software Foundation (ASF) under one or more
% contributor license agreements.  See the NOTICE file distributed with
% this work for additional information regarding copyright ownership.
% The ASF licenses this file to you under the Apache License, Version
% 2.0 (the "License"); you may not use this file except in compliance
% with the License.  You may obtain a copy of the License at
%
%   http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
% implied.  See the License for the specific language governing
% permissions and limitations under the License.

classdef ClassTypeValidator < arrow.array.internal.list.Validator
    
    properties (GetAccess=public, SetAccess=private)
        ClassName(1, 1) string
    end
    
    methods
        function obj = ClassTypeValidator(data)
            obj.ClassName = class(data);
        end
        
        function validateElement(obj, element)
           if ~isa(element, obj.ClassName)
               id = "arrow:array:list:ClassTypeMismatch";
               fmt = "Expected all cell array elements to have class type " + ...
                   """%s"", but encountered an element whose class type is" + ...
                   " ""%s"".";
               msg = compose(fmt, obj.ClassName, class(element));
              error(id, msg);
           end
        end

        function length = getElementLength(~, element)
            length = numel(element);
        end

        function C = reshapeCellElements(~, C)
            C = cellfun(@(elem) reshape(elem, [], 1), C, UniformOutput=false);
        end
    end
end

