#!/usr/bin/env bash
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

set -eux

case "${MINGW_PACKAGE_PREFIX}" in
  mingw-w64-ucrt-x86_64)
    :
    ;;
  *)
    pacman \
      --cascade \
      --noconfirm \
      --nosave \
      --recursive \
      --remove \
      "${MINGW_PACKAGE_PREFIX}-clang-tools-extra" \
      "${MINGW_PACKAGE_PREFIX}-gcc-ada" \
      "${MINGW_PACKAGE_PREFIX}-gcc-fortran" \
      "${MINGW_PACKAGE_PREFIX}-gcc-libgfortran" \
      "${MINGW_PACKAGE_PREFIX}-gcc-objc" \
      "${MINGW_PACKAGE_PREFIX}-libgccjit"
      ;;
esac
