{#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#}
*Extra CI GitHub report for <{{ report.url }}|{{ report.name }}>*
{% if report.failed_jobs() %}
:x: *{{ report.failed_jobs() | length }} failed jobs*
{% for job in report.failed_jobs() -%}
- <{{ job.url }}|{{ job.name }}>
{% endfor %}
{%- endif -%}
{% if report.successful_jobs() %}

:tada: *{{ report.successful_jobs() | length }} successful jobs*
{%- endif -%}

{% if extra_message_success and not report.failed_jobs() %}

{{ extra_message_success }}
{%- endif -%}
{% if extra_message_failure and report.failed_jobs() %}

{{ extra_message_failure }}
{% endif %}