# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestGandivaSelectionVector < Test::Unit::TestCase
  include Helper::Buildable

  def setup
    omit("Gandiva is required") unless defined?(::Gandiva)
  end

  def test_uint16
    selection_vector = Gandiva::UInt16SelectionVector.new(10)
    assert_equal(build_uint16_array([]),
                 selection_vector.to_array)
  end

  def test_uint32
    selection_vector = Gandiva::UInt32SelectionVector.new(10)
    assert_equal(build_uint32_array([]),
                 selection_vector.to_array)
  end

  def test_uint64
    selection_vector = Gandiva::UInt64SelectionVector.new(10)
    assert_equal(build_uint64_array([]),
                 selection_vector.to_array)
  end
end
