# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestMapArray < Test::Unit::TestCase
  include Helper::Buildable

  def setup
    offsets = build_int32_array([0, 2, 5])
    @keys = build_string_array(["a", "b", "c", "d", "e"])
    @items = build_int16_array([0, 1, 2, 3, 4])
    @map_array = Arrow::MapArray.new(offsets,
                                     @keys,
                                     @items)
  end

  def test_keys
    assert_equal(@keys,
                 @map_array.keys)
  end

  def test_items
    assert_equal(@items,
                 @map_array.items)
  end
end
