# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

module Gandiva
  class Loader < GObjectIntrospection::Loader
    class << self
      def load
        super("Gandiva", Gandiva)
      end
    end

    private
    def load_method_info(info, klass, method_name)
      case klass.name
      when "Gandiva::BooleanLiteralNode"
        case method_name
        when "value?"
          method_name = "value"
        end
        super(info, klass, method_name)
      else
        super
      end
    end

    def post_load(repository, namespace)
      require_libraries
    end

    def require_libraries
      require "gandiva/arrow-schema"
      require "gandiva/expression-builder"
    end
  end
end
