# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

module Arrow
  class ArrayStatistics
    if method_defined?(:null_count_exact)
      alias_method :null_count_raw, :null_count
      def null_count
        return nil unless has_null_count?
        if null_count_exact?
          null_count_exact
        else
          null_count_approximate
        end
      end
    end

    if method_defined?(:distinct_count_exact)
      alias_method :distinct_count_raw, :distinct_count
      def distinct_count
        return nil unless has_distinct_count?
        if distinct_count_exact?
          distinct_count_exact
        else
          distinct_count_approximate
        end
      end
    end
  end
end
